use super::Encoder;
use crate::flowgger::config::Config;
use crate::flowgger::record::{Record, SDValue};

#[derive(Clone)]
pub struct LTSVEncoder {
    extra: Vec<(String, String)>,
}

impl LTSVEncoder {
    pub fn new(config: &Config) -> LTSVEncoder {
        let extra = match config.lookup("output.ltsv_extra") {
            None => Vec::new(),
            Some(extra) => extra
                .as_table()
                .expect("output.ltsv_extra must be a list of key/value pairs")
                .iter()
                .map(|(k, v)| {
                    (
                        k.to_owned(),
                        v.as_str()
                            .expect("output.ltsv_extra values must be strings")
                            .to_owned(),
                    )
                })
                .collect(),
        };
        LTSVEncoder { extra }
    }
}

struct LTSVString {
    out: String,
}

impl LTSVString {
    pub fn new() -> LTSVString {
        LTSVString { out: String::new() }
    }

    pub fn insert(&mut self, key: &str, value: &str) {
        if !self.out.is_empty() {
            self.out.push('\t');
        }
        if key.chars().any(|s| s == '\n' || s == '\t' || s == ':') {
            let key_esc = key.replace("\n", " ").replace("\t", " ").replace(":", "_");
            self.out.push_str(&key_esc);
        } else {
            self.out.push_str(key);
        };
        self.out.push(':');
        if value.chars().any(|s| s == '\n' || s == '\t') {
            let value_esc = value.replace("\t", " ").replace("\n", " ");
            self.out.push_str(&value_esc);
        } else {
            self.out.push_str(value);
        };
    }

    pub fn finalize(self) -> String {
        self.out
    }
}

impl Encoder for LTSVEncoder {
    fn encode(&self, record: Record) -> Result<Vec<u8>, &'static str> {
        let mut res = LTSVString::new();
        if let Some(sd) = record.sd {
            for &(ref name, ref value) in &sd.pairs {
                let name = if (*name).starts_with('_') {
                    &name[1..] as &str
                } else {
                    name as &str
                };
                match *value {
                    SDValue::String(ref value) => res.insert(name, value),
                    SDValue::Bool(ref value) => res.insert(name, &value.to_string()),
                    SDValue::F64(ref value) => res.insert(name, &value.to_string()),
                    SDValue::I64(ref value) => res.insert(name, &value.to_string()),
                    SDValue::U64(ref value) => res.insert(name, &value.to_string()),
                    SDValue::Null => res.insert(name, ""),
                }
            }
        }
        for &(ref name, ref value) in &self.extra {
            let name = if (*name).starts_with('_') {
                &name[1..] as &str
            } else {
                name as &str
            };
            res.insert(name, value);
        }
        res.insert("host", &record.hostname);
        res.insert("time", &record.ts.to_string());
        if let Some(msg) = record.msg {
            res.insert("message", &msg);
        }
        if let Some(full_msg) = record.full_msg {
            res.insert("full_message", &full_msg);
        }
        if let Some(severity) = record.severity {
            res.insert("level", &format!("{}", severity));
        }
        if let Some(facility) = record.facility {
            res.insert("facility", &format!("{}", facility));
        }
        if let Some(appname) = record.appname {
            res.insert("appname", &appname);
        }
        if let Some(procid) = record.procid {
            res.insert("procid", &procid);
        }
        if let Some(msgid) = record.msgid {
            res.insert("msgid", &msgid);
        }
        Ok(res.finalize().into_bytes())
    }
}
